@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 This module provides constants used throughout the maestral, the GUI and CLI. It should
 be kept free of memory heavy imports.
@@ -9,7 +8,7 @@ import platform
 from enum import Enum
 
 try:
-    from importlib.resources import path  # type: ignore
+    from importlib.resources import path
 except ImportError:
     from importlib_resources import path  # type: ignore
 
@@ -18,7 +17,7 @@ except ImportError:
 APP_NAME = "Maestral"
 BUNDLE_ID = "com.samschott.maestral"
 APP_ICON_PATH = path("maestral.resources", "maestral.png").__enter__()
-ENV = {"PYTHONOPTIMIZE": "2"}
+ENV = {"PYTHONOPTIMIZE": "2", "LC_CTYPE": "UTF-8"}
 
 # sync
 OLD_REV_FILE = ".maestral"
