@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 Looking at the source code for :class:`watchdog.utils.dirsnapshot.DirectorySnapshotDiff`,
 the event types are categorised as follows:
@@ -60,7 +59,7 @@ into Deleted and Created events by Maestral.
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
-from watchdog.observers.polling import (  # type: ignore
+from watchdog.observers.polling import (
     PollingEmitter,
     PollingObserver,
     FileDeletedEvent,
