@@ -1,7 +1,9 @@
-# -*- coding: utf-8 -*-
 """This module contains the classes to coordinate sync threads."""
 
+from __future__ import annotations
+
 # system imports
+import os
 import errno
 import time
 import gc
@@ -10,13 +12,15 @@ from contextlib import contextmanager
 from functools import wraps
 from queue import Empty, Queue
 from threading import Event, RLock, Thread
-from typing import Iterator, Optional, cast, Dict, List, Type, TypeVar, Callable, Any
+from tempfile import TemporaryDirectory
+from typing import Iterator, cast, TypeVar, Callable, Any
 
 # local imports
 from . import __url__
 from . import notify
-from .client import DropboxClient
-from .config import MaestralConfig
+from .client import DropboxClient, API_HOST
+from .core import TeamRootInfo, UserRootInfo
+from .config import MaestralConfig, MaestralState
 from .constants import (
     DISCONNECTED,
     CONNECTING,
@@ -25,42 +29,46 @@ from .constants import (
     PAUSED,
     CONNECTED,
 )
-from .database import SyncEvent
-from .errors import (
+from .models import SyncEvent
+from .exceptions import (
     CancelledError,
     DropboxConnectionError,
     MaestralApiError,
     InotifyError,
+    NoDropboxDirError,
+    PathRootError,
+    DropboxServerError,
 )
 from .sync import SyncEngine
 from .fsevents import Observer
 from .logging import scoped_logger
-from .utils import exc_info_tuple
+from .utils import removeprefix
 from .utils.integration import check_connection, get_inotify_limits
+from .utils.path import move, delete, is_equal_or_child, is_child, normalize
 
 
 __all__ = ["SyncManager"]
 
+DROPBOX_API_HOSTNAME = "https://" + API_HOST
+
 
 FT = TypeVar("FT", bound=Callable[..., Any])
 
-
-malloc_trim: Optional[Callable]
+malloc_trim: Callable
 
 try:
     libc = ctypes.CDLL("libc.so.6")
     malloc_trim = libc.malloc_trim
 except (OSError, AttributeError):
-    malloc_trim = None
+
+    def malloc_trim(*args):
+        pass
 
 
 def _free_memory() -> None:
     """Give back memory"""
-
     gc.collect()
-
-    if malloc_trim:
-        malloc_trim(0)
+    malloc_trim(0)
 
 
 class SyncManager:
@@ -69,7 +77,7 @@ class SyncManager:
     :param client: The Dropbox API client, a wrapper around the Dropbox Python SDK.
     """
 
-    added_item_queue: "Queue[str]"
+    added_item_queue: Queue[str]
     """Queue of dropbox paths which have been newly included in syncing."""
 
     def __init__(self, client: DropboxClient):
@@ -77,6 +85,7 @@ class SyncManager:
         self.client = client
         self.config_name = self.client.config_name
         self._conf = MaestralConfig(self.config_name)
+        self._state = MaestralState(self.config_name)
         self._logger = scoped_logger(__name__, self.config_name)
 
         self._lock = RLock()
@@ -101,7 +110,7 @@ class SyncManager:
         )
         self.connection_helper.start()
 
-        self.local_observer_thread: Optional[Observer] = None
+        self.local_observer_thread: Observer | None = None
 
     def _with_lock(fn: FT) -> FT:  # type: ignore
         @wraps(fn)
@@ -126,15 +135,15 @@ class SyncManager:
         self._conf.set("sync", "reindex_interval", interval)
 
     @property
-    def activity(self) -> Dict[str, SyncEvent]:
+    def activity(self) -> dict[str, SyncEvent]:
         """Returns a list all items queued for or currently syncing."""
         return self.sync.syncing
 
     @property
-    def history(self) -> List[SyncEvent]:
+    def history(self) -> list[SyncEvent]:
         """A list of the last SyncEvents in our history. History will be kept for the
         interval specified by the config value ``keep_history`` (defaults to two weeks)
-        but at most 1,000 events will kept."""
+        but at most 1,000 events will be kept."""
         return self.sync.history
 
     @property
@@ -159,7 +168,7 @@ class SyncManager:
         if self.running.is_set():
             return
 
-        if not check_connection("www.dropbox.com"):
+        if not check_connection(DROPBOX_API_HOSTNAME, logger=self._logger):
             # Schedule autostart when connection becomes available.
             self.autostart.set()
             self._logger.info(CONNECTING)
@@ -204,7 +213,10 @@ class SyncManager:
                 name="maestral-folder-download",
             )
 
-        if self._conf.get("sync", "upload"):
+        enable_upload = self._conf.get("sync", "upload")
+        enable_download = self._conf.get("sync", "download")
+
+        if enable_upload:
 
             self.upload_thread = Thread(
                 target=self.upload_worker,
@@ -218,16 +230,23 @@ class SyncManager:
             )
 
             if not self.local_observer_thread:
-                self.local_observer_thread = self._create_observer()
+                try:
+                    self.local_observer_thread = self._create_observer()
+                except MaestralApiError as exc:
+                    self._logger.error(exc.title, exc_info=True)
+                    self.sync.desktop_notifier.notify(
+                        exc.title, exc.message, level=notify.ERROR
+                    )
+                    return
 
         self.running.set()
         self.autostart.set()
 
-        if self._conf.get("sync", "upload"):
+        if enable_upload:
             self.sync.fs_events.enable()
             self.upload_thread.start()
 
-        if self._conf.get("sync", "download"):
+        if enable_download:
             self.download_thread.start()
             self.download_thread_added_folder.start()
 
@@ -238,7 +257,7 @@ class SyncManager:
     def _create_observer(self) -> Observer:
 
         local_observer_thread = Observer(timeout=40)
-        local_observer_thread.setName("maestral-fsobserver")
+        local_observer_thread.name = "maestral-fsobserver"
         local_observer_thread.schedule(
             self.sync.fs_events, self.sync.dropbox_path, recursive=True
         )
@@ -251,53 +270,54 @@ class SyncManager:
             local_observer_thread.start()
         except OSError as exc:
 
-            err_cls: Type[MaestralApiError]
-
             if exc.errno in (errno.ENOSPC, errno.EMFILE):
-                title = "Inotify limit reached"
 
                 try:
                     max_user_watches, max_user_instances, _ = get_inotify_limits()
                 except OSError:
-                    max_user_watches, max_user_instances = 2 ** 18, 2 ** 9
+                    max_user_watches, max_user_instances = 2**18, 2**9
 
                 url = f"{__url__}/docs/inotify-limits"
 
                 if exc.errno == errno.ENOSPC:
-                    n_new = max(2 ** 19, 2 * max_user_watches)
+                    n_new = max(2**19, 2 * max_user_watches)
 
-                    msg = (
+                    raise InotifyError(
+                        "Inotify limit reached",
                         "Changes to your Dropbox folder cannot be monitored because it "
                         "contains too many items. Please increase "
                         f"fs.inotify.max_user_watches to {n_new}. See {url} for more "
-                        "information."
+                        "information.",
                     )
 
                 else:
-                    n_new = max(2 ** 10, 2 * max_user_instances)
+                    n_new = max(2**10, 2 * max_user_instances)
 
-                    msg = (
+                    raise InotifyError(
+                        "Inotify limit reached",
                         "Changes to your Dropbox folder cannot be monitored because "
                         "there are too many activity inotify instances. Please "
                         f"increase fs.inotify.max_user_instances to {n_new}. See "
-                        f"{url} for more information."
+                        f"{url} for more information.",
                     )
 
-                err_cls = InotifyError
-
             elif exc.errno in (errno.EPERM, errno.EACCES):
-                title = "Insufficient permissions to monitor local changes"
-                msg = "Please check the permissions for your local Dropbox folder"
-                err_cls = InotifyError
+                raise InotifyError(
+                    "Insufficient permissions to monitor local changes",
+                    "Please check the permissions for your local Dropbox folder",
+                )
 
+            elif exc.errno in (errno.ENOENT, errno.ENOTDIR):
+                raise NoDropboxDirError(
+                    "Dropbox folder missing",
+                    "Please move the Dropbox folder back to its original location "
+                    "or restart Maestral to set up a new folder.",
+                )
             else:
-                title = "Could not start watch of local directory"
-                msg = exc.strerror
-                err_cls = MaestralApiError
-
-            new_error = err_cls(title, msg)
-            self._logger.error(title, exc_info=exc_info_tuple(new_error))
-            self.sync.desktop_notifier.notify(title, msg, level=notify.ERROR)
+                raise MaestralApiError(
+                    "Could not start watch of local directory",
+                    exc.strerror,
+                )
 
         return local_observer_thread
 
@@ -325,7 +345,9 @@ class SyncManager:
         """Resets all saved sync state. Settings are not affected."""
 
         if self.running.is_set():
-            raise RuntimeError("Cannot reset sync state while syncing.")
+            raise MaestralApiError(
+                "Cannot reset sync state while syncing", "Please try again when idle."
+            )
 
         self.sync.reset_sync_state()
 
@@ -345,12 +367,220 @@ class SyncManager:
 
         self.stop()
 
-        self.sync.remote_cursor = ""
-        self.sync.clear_index()
+        self.reset_sync_state()
 
         if was_running:
             self.start()
 
+    # ---- path root management --------------------------------------------------------
+
+    def check_and_update_path_root(self) -> bool:
+        """
+        Checks if the user's root namespace corresponds to the currently configured
+        path root. Updates the root namespace if required and migrates the local
+        folder structure. Syncing will be paused during the migration.
+
+        :returns: Whether the path root was updated.
+        """
+
+        if self._needs_path_root_update():
+
+            was_running = self.running.is_set()
+
+            self.stop()
+
+            self._update_path_root()
+
+            if was_running:
+                self._logger.info("Restarting sync")
+                self.start()
+
+            return True
+        else:
+            self._logger.debug("Path root is up to date")
+            return False
+
+    def _needs_path_root_update(self) -> bool:
+        """
+        Checks if the user's root namespace corresponds to the currently configured
+        path root.
+
+        :returns: Whether the configured root namespace needs to be updated.
+        """
+
+        self._logger.debug("Checking path root...")
+
+        account_info = self.client.get_account_info()
+        return self.client.namespace_id != account_info.root_info.root_namespace_id
+
+    def _update_path_root(self) -> None:
+        """
+        Changes the layout of the local Dropbox folder if the user joins or leaves a
+        team. New team folders will be downloaded, old team folders will be removed.
+        """
+
+        current_root_type = self._state.get("account", "path_root_type")
+        current_user_home_path = self._state.get("account", "home_path")
+        current_user_home_path_lower = normalize(current_user_home_path)
+
+        if current_root_type == "team" and current_user_home_path == "":
+            raise MaestralApiError(
+                "Cannot migrate folder structure",
+                "Inconsistent namespace information found.",
+            )
+
+        root_info = self.client.account_info.root_info
+        team = self.client.account_info.team
+
+        team_name = team.name if team else "team"
+
+        if isinstance(root_info, UserRootInfo):
+            new_root_type = "user"
+            new_user_home_path = ""
+        elif isinstance(root_info, TeamRootInfo):
+            new_root_type = "team"
+            new_user_home_path = root_info.home_path
+        else:
+            raise MaestralApiError(
+                "Unknown root namespace type",
+                f"Got {root_info!r} but expected UserRootInfo or TeamRootInfo.",
+            )
+
+        local_user_home_path = self.sync.to_local_path_from_cased(new_user_home_path)
+
+        try:
+            local_dropbox_dirlist = list(os.scandir(self.sync.dropbox_path))
+        except (FileNotFoundError, NotADirectoryError):
+            title = "Dropbox folder missing"
+            msg = (
+                "Please move the Dropbox folder back to its original location or "
+                "restart Maestral to set up a new folder."
+            )
+            raise NoDropboxDirError(title, msg)
+
+        with self.sync.sync_lock:
+
+            if new_root_type == "team" and current_root_type == "user":
+
+                # User joined a team.
+
+                self._logger.info("User joined %s. Resyncing user files.", team_name)
+
+                self.sync.desktop_notifier.notify(
+                    f"Joined {team_name}",
+                    "Migrating user files and downloading team folders",
+                )
+
+                # Migrate user folder to "self.sync.dropbox_path/home_path". We do this
+                # by creating a temporary folder and renaming it after moving all
+                # personal items to prevent name conflicts where a folder has the same
+                # name as `root_info.home_path`.
+
+                tmpdir = TemporaryDirectory(dir=self.sync.dropbox_path)
+
+                for entry in local_dropbox_dirlist:
+                    if entry.path != tmpdir.name:
+                        new_path = f"{tmpdir.name}/{entry.name}"
+                        self._logger.debug(
+                            "Moving to personal folder: %r → %r", entry.path, new_path
+                        )
+                        move(entry.path, new_path, raise_error=True)
+
+                self._logger.debug("Moving %r → %r", tmpdir.name, local_user_home_path)
+                os.rename(tmpdir.name, local_user_home_path)
+
+                # Migrate all excluded items.
+                self._logger.debug("Migrating excluded items")
+                new_excluded = [
+                    new_user_home_path + path for path in self.sync.excluded_items
+                ]
+                self.sync.excluded_items = new_excluded
+
+            elif new_root_type == "user" and current_root_type == "team":
+
+                # User left a team.
+
+                self._logger.info("User left team. Updating folder layout.")
+
+                self.sync.desktop_notifier.notify(
+                    "Left Dropbox Team",
+                    "Migrating user files and removing team folders",
+                )
+
+                # Remove all team folders.
+                for entry in local_dropbox_dirlist:
+                    if entry.name != current_user_home_path.lstrip("/"):
+                        delete(entry.path, raise_error=True)
+
+                # Migrate user folders to local Dropbox root. We do this by renaming the
+                # user home to a temporary name and then moving its contents to the
+                # parent folder.
+
+                old_home_root = self.sync.dropbox_path + current_user_home_path
+
+                tmpdir = TemporaryDirectory(dir=self.sync.dropbox_path)
+
+                try:
+                    os.rename(old_home_root, tmpdir.name)
+                except (FileNotFoundError, NotADirectoryError):
+                    # User folder does not (yet) exist.
+                    pass
+                else:
+                    for entry in list(os.scandir(tmpdir.name)):
+                        new_path = f"{self.sync.dropbox_path}/{entry.name}"
+                        move(entry.path, new_path, raise_error=True)
+                        self._logger.debug(
+                            "Moved to root folder: %r → %r", entry.path, new_path
+                        )
+
+                delete(tmpdir.name)
+
+                # Migrate excluded items:
+                # Prune all teams folders from excluded list. Remove home folder
+                # prefix from excluded items. If the user folder itself is
+                # excluded, keep it excluded.
+                self._logger.debug("Migrating excluded items")
+                new_excluded = [
+                    removeprefix(path, current_user_home_path_lower)
+                    for path in self.sync.excluded_items
+                    if is_child(path, current_user_home_path_lower)
+                ]
+
+                self.sync.excluded_items = new_excluded
+
+            elif new_root_type == "team" and current_root_type == "team":
+
+                # User switched between different teams.
+
+                self._logger.info("User switched teams. Updating team folders.")
+
+                self.sync.desktop_notifier.notify(
+                    f"Switched teams to {team_name}", "Updating team folders"
+                )
+
+                # Remove all team folders, leave user folder alone.
+                for entry in local_dropbox_dirlist:
+                    if entry.name != current_user_home_path.lstrip("/"):
+                        delete(entry.path, raise_error=True)
+                        self._logger.debug("Deleted team folder: %r", entry.path)
+
+                # Migrate excluded items:
+                # Prune all teams folders from excluded list. If the user folder
+                # itself is excluded, keep it excluded.
+                self._logger.debug("Migrating excluded items")
+                new_excluded = [
+                    path
+                    for path in self.sync.excluded_items
+                    if is_equal_or_child(path, current_user_home_path_lower)
+                ]
+                self.sync.excluded_items = new_excluded
+
+            # Update path root of client.
+            self.client.update_path_root(root_info)
+
+            #  Trigger reindex.
+            self.sync.reset_sync_state()
+
     # ---- thread methods --------------------------------------------------------------
 
     def connection_monitor(self) -> None:
@@ -362,7 +592,7 @@ class SyncManager:
 
         while self._connection_helper_running:
 
-            connected = check_connection("www.dropbox.com")
+            connected = check_connection(DROPBOX_API_HOSTNAME)
 
             if connected != self.connected:
                 # Log the status change.
@@ -385,7 +615,7 @@ class SyncManager:
         """
         Worker to sync changes of remote Dropbox with local folder.
 
-        :param running: Event to shutdown local file event handler and worker threads.
+        :param running: Event to shut down local file event handler and worker threads.
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
@@ -402,6 +632,11 @@ class SyncManager:
                         self.sync.remote_cursor, client=client
                     )
 
+                    # Check for root namespace updates. Don't apply any remote
+                    # changes in case of a changed root path.
+                    if self.check_and_update_path_root():
+                        return
+
                     if not running.is_set():
                         return
 
@@ -412,7 +647,7 @@ class SyncManager:
                         self.sync.download_sync_cycle(client)
                         self._logger.info(IDLE)
 
-                        client.get_space_usage()  # update space usage
+                        client.get_space_usage()
 
         _free_memory()
 
@@ -425,7 +660,7 @@ class SyncManager:
         """
         Worker to download items which have been newly included in sync.
 
-        :param running: Event to shutdown local file event handler and worker threads.
+        :param running: Event to shut down local file event handler and worker threads.
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
@@ -437,12 +672,12 @@ class SyncManager:
             with self._handle_sync_thread_errors(running, autostart):
 
                 try:
-                    dbx_path = self.added_item_queue.get(timeout=40)
+                    dbx_path_lower = self.added_item_queue.get(timeout=40)
                 except Empty:
                     pass
                 else:
-                    # protect against crashes
-                    self.sync.pending_downloads.add(dbx_path.lower())
+                    # Guard against crashes.
+                    self.sync.pending_downloads.add(dbx_path_lower)
 
                     if not running.is_set():
                         return
@@ -450,9 +685,9 @@ class SyncManager:
                     with self.sync.sync_lock:
 
                         with self.sync.client.clone_with_new_session() as client:
-                            self.sync.get_remote_item(dbx_path, client)
+                            self.sync.get_remote_item(dbx_path_lower, client)
 
-                        self.sync.pending_downloads.discard(dbx_path)
+                        self.sync.pending_downloads.discard(dbx_path_lower)
 
                         self._logger.info(IDLE)
 
@@ -467,7 +702,7 @@ class SyncManager:
         """
         Worker to sync local changes to remote Dropbox.
 
-        :param running: Event to shutdown local file event handler and worker threads.
+        :param running: Event to shut down local file event handler and worker threads.
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
@@ -501,24 +736,37 @@ class SyncManager:
         """
         Worker to sync local changes to remote Dropbox.
 
-        :param running: Event to shutdown local file event handler and worker threads.
+        :param running: Event to shut down local file event handler and worker threads.
         :param startup_completed: Set when startup sync is completed.
         :param autostart: Set when syncing should automatically resume on connection.
         """
 
         with self._handle_sync_thread_errors(running, autostart):
 
+            # Fail early if Dropbox folder disappeared.
+            self.sync.ensure_dropbox_folder_present()
+
             # Reload mignore rules.
             self.sync.load_mignore_file()
 
+            self.client.get_space_usage()
+
+            # Update path root and migrate local folders. This is required when a user
+            # joins or leaves a team and their root namespace changes.
+            self.check_and_update_path_root()
+
+            if not running.is_set():
+                startup_completed.set()
+                return
+
             with self.sync.client.clone_with_new_session() as client:
 
                 # Retry failed downloads.
                 if len(self.sync.download_errors) > 0:
                     self._logger.info("Retrying failed syncs...")
 
-                for dbx_path in list(self.sync.download_errors):
-                    self.sync.get_remote_item(dbx_path, client)
+                for error in list(self.sync.download_errors):
+                    self.sync.get_remote_item(error.dbx_path_lower, client)
 
                 # Resume interrupted downloads.
                 if len(self.sync.pending_downloads) > 0:
@@ -556,20 +804,23 @@ class SyncManager:
         try:
             yield
         except CancelledError:
+            # Shutdown will be handled externally.
             running.clear()
-        except DropboxConnectionError:
-            self._logger.info(DISCONNECTED)
+        except (DropboxConnectionError, DropboxServerError):
             self._logger.debug("Connection error", exc_info=True)
-            running.clear()
-            autostart.set()
+            self._logger.info(DISCONNECTED)
+            self.stop()
+            self.autostart.set()
             self._logger.info(CONNECTING)
+        except PathRootError:
+            self._logger.debug("API call failed due to path root error", exc_info=True)
+            self.check_and_update_path_root()
         except Exception as err:
-            running.clear()
-            autostart.clear()
             title = getattr(err, "title", "Unexpected error")
             message = getattr(err, "message", "Please restart to continue syncing")
             self._logger.error(title, exc_info=True)
             self.sync.desktop_notifier.notify(title, message, level=notify.ERROR)
+            self.stop()
 
     def __del__(self):
         try:
