@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
 This module contains functions for common path operations.
 """
@@ -14,7 +13,7 @@ from stat import S_ISDIR
 from typing import List, Optional, Tuple, Callable, Iterator, Iterable, Union
 
 # local imports
-from .content_hasher import DropboxContentHasher
+from .hashing import DropboxContentHasher
 
 
 def _path_components(path: str) -> List[str]:
@@ -23,6 +22,44 @@ def _path_components(path: str) -> List[str]:
     return cleaned_components
 
 
+Path = Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"]
+
+# ==== path relationships ==============================================================
+
+
+def is_child(path: str, parent: str) -> bool:
+    """
+    Checks if ``path`` semantically is inside ``parent``. Neither path needs to
+    refer to an actual item on the drive. This function is case-sensitive.
+
+    :param path: Item path.
+    :param parent: Parent path.
+    :returns: Whether ``path`` semantically lies inside ``parent``.
+    """
+
+    parent = parent.rstrip(osp.sep) + osp.sep
+    path = path.rstrip(osp.sep)
+
+    return path.startswith(parent)
+
+
+def is_equal_or_child(path: str, parent: str) -> bool:
+    """
+    Checks if ``path`` semantically is inside ``parent`` or equals ``parent``. Neither
+    path needs to refer to an actual item on the drive. This function is case-sensitive.
+
+    :param path: Item path.
+    :param parent: Parent path.
+    :returns: ``True`` if ``path`` semantically lies inside ``parent`` or
+        ``path == parent``.
+    """
+
+    return is_child(path, parent) or path == parent
+
+
+# ==== case sensitivity and normalization ==============================================
+
+
 def normalize_case(string: str) -> str:
     """
     Converts a string to lower case. Todo: Follow Python 2.5 / Dropbox conventions.
@@ -77,47 +114,21 @@ def is_fs_case_sensitive(path: str) -> bool:
     :param path: Path to check.
     :returns: Whether ``path`` lies on a partition with a case-sensitive file system.
     """
+
+    if path == osp.pathsep:
+        raise ValueError(f"Cannot check '{osp.pathsep}'")
+
     if path.islower():
         check_path = path.upper()
     else:
         check_path = path.lower()
 
-    if osp.exists(path) and not osp.exists(check_path):
+    if exists(path) and not exists(check_path):
         return True
     else:
         return not osp.samefile(path, check_path)
 
 
-def is_child(path: str, parent: str) -> bool:
-    """
-    Checks if ``path`` semantically is inside ``parent``. Neither path needs to
-    refer to an actual item on the drive. This function is case sensitive.
-
-    :param path: Item path.
-    :param parent: Parent path.
-    :returns: Whether ``path`` semantically lies inside ``parent``.
-    """
-
-    parent = parent.rstrip(osp.sep) + osp.sep
-    path = path.rstrip(osp.sep)
-
-    return path.startswith(parent)
-
-
-def is_equal_or_child(path: str, parent: str) -> bool:
-    """
-    Checks if ``path`` semantically is inside ``parent`` or equals ``parent``. Neither
-    path needs to refer to an actual item on the drive. This function is case sensitive.
-
-    :param path: Item path.
-    :param parent: Parent path.
-    :returns: ``True`` if ``path`` semantically lies inside ``parent`` or
-        ``path == parent``.
-    """
-
-    return is_child(path, parent) or path == parent
-
-
 def equivalent_path_candidates(
     path: str,
     root: str = osp.sep,
@@ -128,7 +139,7 @@ def equivalent_path_candidates(
     function, this method returns a list of matching un-normalized local paths.
 
     If no such local path exists, the normalized path itself is returned. If a local
-    path can be followed up to a certain parent in the hierarchy, it will be taked and
+    path can be followed up to a certain parent in the hierarchy, it will be taken and
     the remaining normalized path will be appended.
 
     :Example:
@@ -227,7 +238,7 @@ def to_existing_unnormalized_path(path: str, root: str = osp.sep) -> str:
     candidates = equivalent_path_candidates(path, root)
 
     for candidate in candidates:
-        if osp.exists(candidate):
+        if exists(candidate):
             return candidate
 
     raise FileNotFoundError(f'No matches with different casing found in "{root}"')
@@ -248,7 +259,7 @@ def normalized_path_exists(path: str, root: str = osp.sep) -> bool:
     candidates = equivalent_path_candidates(path, root)
 
     for c in candidates:
-        if osp.exists(c):
+        if exists(c):
             return True
 
     return False
@@ -257,7 +268,7 @@ def normalized_path_exists(path: str, root: str = osp.sep) -> bool:
 def generate_cc_name(path: str, suffix: str = "conflicting copy") -> str:
     """
     Generates a path for a conflicting copy of ``path``. The file name is created by
-    inserting the given ``suffix`` between the the filename and extension. For instance:
+    inserting the given ``suffix`` between the filename and the extension. For instance:
 
         "my_file.txt" -> "my_file (conflicting copy).txt"
 
@@ -284,9 +295,12 @@ def generate_cc_name(path: str, suffix: str = "conflicting copy") -> str:
     return osp.join(dirname, cc_candidate)
 
 
+# ==== higher level file operations ====================================================
+
+
 def delete(path: str, raise_error: bool = False) -> Optional[OSError]:
     """
-    Deletes a file or folder at ``path``.
+    Deletes a file or folder at ``path``. Symlinks will not be followed.
 
     :param path: Path of item to delete.
     :param raise_error: Whether to raise errors or return them.
@@ -295,7 +309,7 @@ def delete(path: str, raise_error: bool = False) -> Optional[OSError]:
     err = None
 
     try:
-        shutil.rmtree(path)
+        shutil.rmtree(path)  # Will raise OSError when it finds a symlink.
     except OSError:
         try:
             os.unlink(path)
@@ -335,7 +349,7 @@ def move(
     if preserve_dest_permissions:
         # save dest permissions
         try:
-            orig_mode = os.stat(dest_path).st_mode & 0o777
+            orig_mode = os.stat(dest_path, follow_symlinks=False).st_mode & 0o777
         except FileNotFoundError:
             pass
 
@@ -362,7 +376,9 @@ def move(
 
 def walk(
     root: Union[str, os.PathLike],
-    listdir: Callable[[Union[str, os.PathLike]], Iterable[os.DirEntry]] = os.scandir,
+    listdir: Callable[
+        [Union[str, "os.PathLike[str]"]], Iterable[os.DirEntry]
+    ] = os.scandir,
 ) -> Iterator[Tuple[str, os.stat_result]]:
     """
     Iterates recursively over the content of a folder.
@@ -376,33 +392,36 @@ def walk(
 
         try:
             path = entry.path
-            stat = entry.stat()
+            stat = entry.stat(follow_symlinks=False)
 
             yield path, stat
 
             if S_ISDIR(stat.st_mode):
-                for res in walk(entry.path, listdir=listdir):
-                    yield res
+                yield from walk(entry.path, listdir=listdir)
 
         except OSError as exc:
             # Directory may have been deleted between finding it in the directory
             # list of its parent and trying to list its contents. If this
-            # happens we treat it as empty. Likewise if the directory was replaced
-            # with a file of the same name (less likely, but possible).
+            # happens we treat it as empty. Likewise, if the directory was replaced
+            # with a file of the same name (less likely, but possible), it will be
+            # treated as empty.
             if exc.errno in (errno.ENOENT, errno.ENOTDIR, errno.EINVAL):
                 return
             else:
                 raise
 
 
+# ==== miscellaneous utilities =========================================================
+
+
 def content_hash(
-    local_path: str, chunk_size: int = 1024
+    local_path: str, chunk_size: int = 65536
 ) -> Tuple[Optional[str], Optional[float]]:
     """
     Computes content hash of a local file.
 
     :param local_path: Absolute path on local drive.
-    :param chunk_size: Size of chunks to hash in bites.
+    :param chunk_size: Size of chunks to hash in bytes.
     :returns: Content hash to compare with Dropbox's content hash and mtime just before
         the hash was computed.
     """
@@ -410,10 +429,10 @@ def content_hash(
     hasher = DropboxContentHasher()
 
     try:
-        mtime = os.stat(local_path).st_mtime
+        mtime = os.stat(local_path, follow_symlinks=False).st_mtime
 
         try:
-            with open(local_path, "rb") as f:
+            with open(local_path, "rb", opener=opener_no_symlink) as f:
                 while True:
                     chunk = f.read(chunk_size)
                     if len(chunk) == 0:
@@ -422,8 +441,14 @@ def content_hash(
 
         except IsADirectoryError:
             return "folder", mtime
-        else:
-            return str(hasher.hexdigest()), mtime
+
+        except OSError as exc:
+            if exc.errno == errno.ELOOP:
+                hasher.update(b"")  # use empty file for symlinks
+            else:
+                raise exc
+
+        return str(hasher.hexdigest()), mtime
 
     except FileNotFoundError:
         return None, None
@@ -432,3 +457,111 @@ def content_hash(
         return None, None
     finally:
         del hasher
+
+
+def fs_max_lengths_for_path(path: str = "/") -> Tuple[int, int]:
+    """
+    Return the maximum length of file names and paths allowed on a file system.
+
+    :param path: Path to check. This can be specified because different paths may be
+        residing on different file systems. If the given path does not exist, the first
+        existing parent directory in the tree be taken.
+    :returns: Tuple giving the maximum file name and total path lengths.
+    """
+
+    path = osp.abspath(path)
+    dirname = osp.dirname(path)
+
+    while True:
+        try:
+            max_char_path = os.pathconf(dirname, "PC_PATH_MAX")
+            max_char_name = os.pathconf(dirname, "PC_NAME_MAX")
+            return max_char_path, max_char_name
+        except (FileNotFoundError, NotADirectoryError):
+            dirname = osp.dirname(dirname)
+        except ValueError:
+            raise RuntimeError("Cannot get file length limits.")
+        except OSError as exc:
+            if exc.errno == errno.EINVAL:
+                raise RuntimeError("Cannot get file length limits.")
+            else:
+                dirname = "/"
+
+
+# ==== symlink-proof os methods ========================================================
+
+
+def opener_no_symlink(path: Path, flags: int) -> int:
+    """
+    Opener that does not follow symlinks. Uses :meth:`os.open` under the hood.
+
+    :param path: Path to open.
+    :param flags: Flags passed to :meth:`os.open`. O_NOFOLLOW will be added.
+    :return: Open file descriptor.
+    """
+    flags |= os.O_NOFOLLOW
+    return os.open(path, flags=flags)
+
+
+def _get_stats_no_symlink(path: Path) -> Optional[os.stat_result]:
+    try:
+        return os.stat(path, follow_symlinks=False)
+    except (FileNotFoundError, NotADirectoryError):
+        return None
+
+
+def exists(path: Path) -> bool:
+    """Returns whether an item exists at the path. Returns True for symlinks."""
+    return _get_stats_no_symlink(path) is not None
+
+
+def isfile(path: Path) -> bool:
+    """Returns whether a file exists at the path. Returns True for symlinks."""
+    stat = _get_stats_no_symlink(path)
+
+    if stat is None:
+        return False
+    else:
+        return not S_ISDIR(stat.st_mode)
+
+
+def isdir(path: Path) -> bool:
+    """Returns whether a folder exists at the path. Returns False for symlinks."""
+    stat = _get_stats_no_symlink(path)
+
+    if stat is None:
+        return False
+    else:
+        return S_ISDIR(stat.st_mode)
+
+
+def getsize(path: Path) -> int:
+    """Returns the size. Returns False for symlinks."""
+    stat = os.stat(path, follow_symlinks=False)
+    return stat.st_size
+
+
+def get_symlink_target(local_path: str) -> Optional[str]:
+    """
+    Returns the symlink target of a file.
+
+    :param local_path: Absolute path on local drive.
+    :returns: Symlink target of local file. None if the local path does not refer to
+        a symlink or does not exist.
+    """
+
+    try:
+        return os.readlink(local_path)
+    except (FileNotFoundError, NotADirectoryError):
+        return None
+    except OSError as err:
+
+        if err.errno == errno.EINVAL:
+            # File is not a symlink.
+            return None
+
+        if err.errno == errno.ENAMETOOLONG:
+            # Path cannot exist on this filesystem.
+            return None
+
+        raise err
